﻿using System;
using System.Collections;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Presentation;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Presentation.Media;
using Microsoft.SPOT.Presentation.Shapes;
using Microsoft.SPOT.Touch;

using Microsoft.SPOT.Net.NetworkInformation;
using Microsoft.SPOT.Time;

using Gadgeteer.Networking;
using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;
using Gadgeteer.Modules.GHIElectronics;

using Microsoft.WindowsAzure.MobileServices;
using Microsoft.Azure.Zumo.MicroFramework.Core;

namespace TempHumid
{
    public partial class Program
    {
        private const bool debug = true; // control debug display

        private MobileServiceClient mobileService;
        private static readonly string WAMSAppKey = "WAMS-APP-KEY";
        private static readonly string WAMSUrl = "http://iot-kit.azure-mobile.net/";
        private static readonly string TableName = "temphumid";
        private static readonly Guid SensorDeviceId = new Guid(
            // {E0775C30-F449-46C4-A7A6-ECE711F5CFD3}
            0xe0775c30, 0xf449, 0x46c4, 0xa7, 0xa6, 0xec, 0xe7, 0x11, 0xf5, 0xcf, 0xd3);

        private string ipAddress = "";
        private bool isNetworkConnected = false;
        private bool isAzureOK = false;

        private const int timeSec = 10;
        private const int timerMS = timeSec * 1000;
        private static byte[] TimeServerIPAddress = new byte[] { 133, 243, 238, 163 }; //ntp-b2.nict.jp

        private double temperature;
        private double humidity;
        private const double temperatureWarningThreshold = 28.0;
        private const double humidityWarningThreshold = 70.0;

        // This method is run when the mainboard is powered up or reset.   
        void ProgramStarted()
        {
            // Use Debug.Print to show messages in Visual Studio's "Output" window during debugging.
            Debug.Print("Program Started, interval = " + timeSec + " secs");

            SetupNetwork();
            if (isNetworkConnected)
            {
                SyncTimeService(); // Get time of day
                Thread.Sleep(500); // Wait for time event
                if (timeSynced)
                {
                    Debug.Print("Time synced: " + DateTime.Now);
                }
                else
                {
                    Debug.Print("Warning! Time is not synced");
                    multicolorLED.FadeRepeatedly(GT.Color.Yellow);
                }
            }
            else
            {
                Debug.Print("No network connected. Running on local mode.");
            }

            SetupSensors();
            if (isNetworkConnected)
            {
                SetupMobileService();
            }
            else
            {
                multicolorLED.TurnWhite(); // Network error
            }
        }

        private void SetupNetwork()
        {
            string saveIpAddress = "";
            string saveSubnetMask = "";
            string saveGatewayAddress = "";
            string[] saveDnsAddresses = null;
            GHI.Networking.EthernetBuiltIn ei = ethernetJ11D.NetworkInterface;
            NetworkInterface en = ethernetJ11D.NetworkSettings;

            Thread.Sleep(500);
            if (!ei.Opened)
            {
                Debug.Print("now open J11D");
                ei.Open();
            }
            if (ei.IsDhcpEnabled)
            {
                Debug.Print("now enable DHCP");
                ei.EnableDhcp();
            }
            else
            {
                // Save Flash ROM settings
                char[] caIpAddress = en.IPAddress.ToCharArray();
                saveIpAddress = new string(caIpAddress);

                char[] caSubnetMask = en.SubnetMask.ToCharArray();
                saveSubnetMask = new string(caSubnetMask);

                char[] caGatewayAddress = en.GatewayAddress.ToCharArray();
                saveGatewayAddress = new string(caGatewayAddress);

                int numOfDns = en.DnsAddresses.Length;
                saveDnsAddresses = new string[numOfDns];
                if (numOfDns > 0)
                {
                    char[] dnsAddress0 = en.DnsAddresses[0].ToCharArray();
                    saveDnsAddresses[0] = new string(dnsAddress0);
                }
                if (numOfDns > 1)
                {
                    char[] dnsAddress1 = en.DnsAddresses[1].ToCharArray();
                    saveDnsAddresses[1] = new string(dnsAddress1);
                }
            }

            PrintNetworkInfo("Start settings");
            foreach (var ni in NetworkInterface.GetAllNetworkInterfaces())
            {
                if (ni.NetworkInterfaceType == NetworkInterfaceType.Ethernet)
                {
                    if (ni.IsDhcpEnabled)
                    {
                        ipAddress = ni.IPAddress;
                        if (ipAddress != null && ipAddress != "0.0.0.0")
                        {
                            break;
                        }
                        else
                        {
                            ni.RenewDhcpLease();
                            Thread.Sleep(500);
                            ipAddress = ni.IPAddress;
                        }
                    }
                    else
                    {
                        ni.EnableDhcp();
                        Thread.Sleep(500);
                        ni.EnableStaticIP(saveIpAddress, saveSubnetMask, saveGatewayAddress);
                        ni.EnableStaticDns(saveDnsAddresses);
                        ipAddress = ni.IPAddress;
                    }
                    break;
                }
            }
            isNetworkConnected = ipAddress != null && ipAddress != "0.0.0.0";
            PrintNetworkInfo("Completed");
        }

        void PrintNetworkInfo(string arg)
        {
            NetworkInterface en = ethernetJ11D.NetworkSettings;
            const string line = "-----------------";
            Debug.Print(line + " " + arg + " " + line);
            Debug.Print("IP Address:   " + en.IPAddress);
            Debug.Print("DHCP Enabled: " + en.IsDhcpEnabled);
            Debug.Print("Subnet Mask:  " + en.SubnetMask);
            Debug.Print("Gateway:      " + en.GatewayAddress);
            Debug.Print(line + line + line);
        }

        private void SyncTimeService()
        {
            TimeService.SystemTimeChanged += TimeService_SystemTimeChanged;
            TimeServiceStatus status = TimeService.UpdateNow(TimeServerIPAddress, 10);
            TimeService.SetTimeZoneOffset(540); // JST time origin 9h
        }

        bool timeSynced = false;
        void TimeService_SystemTimeChanged(object sender, SystemTimeChangedEventArgs e)
        {
            lock (this)
            {
                timeSynced = true;
            }
        }

        private void SetupSensors()
        {
            GT.Timer timer = new GT.Timer(timerMS);
            timer.Tick += TempHumidSI70_Measurement;
            timer.Behavior = GT.Timer.BehaviorType.RunContinuously;
            timer.Start();
        }


        private void SetupMobileService()
        {
            mobileService = new MobileServiceClient(new Uri(WAMSUrl), WAMSAppKey);
        }

        void TempHumidSI70_Measurement(GT.Timer timer)
        {
            var reading = new SensorReading()
            {
                SensorDeviceId = SensorDeviceId.ToString(),
                MeasuredDate = DateTime.Now,
                Alert = 0
            };

            TempHumidSI70.Measurement measurement = tempHumidSI70.TakeMeasurement();
            this.temperature = measurement.Temperature;
            this.humidity = measurement.RelativeHumidity;

            if (debug)
                Debug.Print("SI70: " + measurement);

            reading.Temperature = temperature;
            reading.Humidity = humidity;

            if (reading.Temperature >= temperatureWarningThreshold)
            {
                reading.Alert |= 1;
            }
            if (reading.Humidity >= humidityWarningThreshold)
            {
                reading.Alert |= 2;
            }

            BeginInvoke(
                    new DispatcherOperationCallback(LedControl), reading.Alert.ToString());

            if (debug)
            {
                Debug.Print("Date:" + reading.MeasuredDate + ", "
                    + temperature.ToString("F1") + "℃, "
                    + humidity.ToString("F1") + "%, "
                    + reading.Alert);
            }

            try
            {
                GC.WaitForPendingFinalizers();
                var response = mobileService.GetTable(TableName).Insert(reading);
                if (response == null)
                {
                    Debug.Print("Warning! No response from Mobile Service.");
                    isAzureOK = false;
                }
                else
                {
                    isAzureOK = true;
                }
                GC.WaitForPendingFinalizers();
            }
            catch (Exception ex)
            {
                Debug.Print(ex.Message);
                isAzureOK = false;
            }

            if (debug)
            {
                Debug.Print("Azure status:" + isAzureOK);
            }

            if (!isAzureOK)
            {
                multicolorLED.TurnWhite();
            }
        }

        object LedControl(object o)
        {
            string alert = o as string;
            switch (alert)
            {
                case ("1"):
                    multicolorLED.TurnRed();
                    break;
                case ("2"):
                    multicolorLED.TurnBlue();
                    break;
                case ("3"):
                    multicolorLED.TurnColor(GT.Color.Purple);
                    break;
                default:
                    multicolorLED.TurnGreen();
                    break;
            }
            return null;
        }

        public class SensorReading : IMobileServiceEntity
        {
            public int Id { get; set; }
            public string SensorDeviceId { get; set; }
            public double Temperature { get; set; }
            public double Humidity { get; set; }
            public int Alert { get; set; }
            public DateTime MeasuredDate { get; set; }
        }
    }
}

